from lark import Transformer, Lark
from .lang import SetVar, RawToken, Op, Statement, Expression, Block, Dotpath, Pointer, Command, Switch, Dummy
import codecs

class StretchBuilder(Transformer):
    def start(self, code):
        init_statements = {}
        statements = {}
        line = 0
        for stat, init in [st for st in code if st is not None]:
            if init:
                init_statements[line] = stat
            else:
                statements[line] = stat
            line += 1
        return Block(statements, init_statements, line)
    def statement(self, stat):
        return [part for part in stat if part is not None], False
    def init_statement(self, stat):
        return [part for part in stat if part is not None], True
    def section(self, section):
        return section
    def part(self, pt):
        return pt[0]
    
    def NEWLINE(self, nwl):
        return None
    def l_brack(self, brack):
        return None
    def r_brack(self, brack):
        return None    
    
    def init(self, init):
        return None
    def COMMA(self, comma):
        return None
    def semi(self, semi):
        return None
    

    def atom(self, atom):
        return atom[0]
    def term(self, term):
        while isinstance(term, list):
            term = term[0]
        return term
    def par_term(self, term):
        return term[1:-1]
    def expr(self, expr):
        if len(expr) == 1:
            return expr[0]
        return Expression(expr)
    
    def block(self, block):
        return block[1]

    
    def STRING(self, str_tok):
        inner = str_tok[1:-1] # remove quotes return
        return codecs.decode(inner, "unicode_escape")
    def string_content(self, content):
        return content
    def STRING_TEXT(self, text):
        return str(text)
    def brace_group(self, group):
        if isinstance(group[0], list):
            group = group[0]
        return '"' + "".join(group) + '"'
    def brack_group(self, group):
        if isinstance(group[0], list):
            group = group[0]
        return "'" + "".join(group) + "'"
    
    def NUM(self, num_tok):
        return float(num_tok) if '.' in num_tok else int(num_tok)
    def NAME(self, token):
        return RawToken(str(token))
    def access(self, acc):
        return None
    def MARK(self, mark):
        return True
    
    
    def switch(self, switch):
        return Switch(switch[0])
    def command(self, cmd):
        return Command(cmd[-1].literal, *[switch.literal for switch in cmd[:-1] if switch is not None])
    
    def name(self, path):
        if len(path) > 1:
            return Dotpath(*[seg for seg in path if seg is not None])
        return path[0]
    def pointer(self, point):
        mark, reference = point
        return Pointer(reference, mark or False)
    
    def dummy(self, dummy):
        return Dummy(dummy[0])
    
    OPERATOR = Op
    def set(self, set_tok):
        return SetVar(set_tok[0])
    def group(self, group):
        return tuple([item for item in group if item is not None])
        

grammar = r"""
    
    
    OPERATOR: OP+ OP_CHAR* OP* | OP* OP_CHAR* OP+
    
    OP: /[+\-*\/=<>!&%^~]/
    OP_CHAR: /[A-Za-z0-9_]/

    colon : ":"
    semi : ";"
    access : "."
    l_access : "<"
    r_access : ">"
    init : "$"
    l_par : "("
    r_par : ")"
    COMMA : ","
    l_brack : "{"
    r_brack : "}"
    MARK : "@"
        
    start: (statement semi | init_statement semi)*
    statement: [set] expr+
    init_statement: init [set] expr+
    set: (name | group) ":"
    
    block: l_brack start r_brack
    
    newline: NEWLINE*
    
    atom: STRING
    | NUM
    | name
    | command
    | pointer
    | block
    | group
    | dummy
    
    dummy: "<-" expr
    
    switch:  "@" NAME
    command: switch* (("(" NAME ")") | (NAME "->"))
    
    group: atom (COMMA atom)+

    par_term: (l_par expr r_par)
    term: atom
    | par_term
    expr: term (OPERATOR term)*
    
    pointer: "[" [MARK] expr "]"
    name: NAME (access NAME)*
    
    COMMENT: /\#[^\n]*/
    %ignore COMMENT
    %import common.NEWLINE
    %import common.CNAME -> NAME
    %import common.ESCAPED_STRING -> STRING

    %import common.SIGNED_NUMBER -> NUM
    
    %import common.WS
    %ignore WS


"""

parser = Lark(grammar, parser="lalr", transformer=StretchBuilder())



from .lang import Dotpath
def run_stretch(path):
    path = Dotpath(*path.split('.'))
    
    from stretch.parser import parser
    from stretch.interpreter import Interpreter

    interpreter = Interpreter()

    with open(path.file('str'), 'r') as test_file:
        main = parser.parse(test_file.read())
        
        interpreter.process_scope(main)
        interpreter.init_scope(main) # because of how scopes are added, must init second in this case

    interpreter.mainloop()


import sys
from . import run_stretch

def run():
    if len(sys.argv) > 1:
        global file
        run_stretch(sys.argv[1])
    else:
        print("Please enter the name of a stretch file to interpret it.")


from .scope_types import View, InitView
from .core_types import Stack, Channel


class Interpreter:
    def __init__(self, parent=None):
        self.parent = parent
        self.__modules__ = {}
        self.__extensions__ = {}
        self.__builtin__ = {}
    
        self.view_stack = Stack()
        self.try_stack = Stack()
        self.exception_stack = Stack()
        self.running = False
        
        self.stack = Stack()
        self.channels = Channel()
    
    def init_scope(self, scope):
        view = InitView(scope)
        self.view_stack.insert(view)
    
    def process_scope(self, scope):
        view = View(scope)
        self.view_stack.insert(view)
    
    def run_scope(self, scope):
        initview = InitView(scope)
        view = View(scope)
        self.view_stack.insert(view)
        self.view_stack.insert(initview)
    
    def inline_scope(self, scope):
        inline_interpreter = Interpreter(self)
        inline_interpreter.run_scope(scope)
        inline_interpreter.mainloop()
    
    
    
    def mainloop(self):
        self.running = True

        while self.running:
            if len(self.view_stack) > 2000:
                raise RecursionError()
            view = self.view_stack.peek()
            if view is None:
                self.running = False
                return
        
            try:
                if not view.step(self):
                    view = self.view_stack.pull()
            except Exception as e:
                if len(self.try_stack) > 0:
                    try_view = self.try_stack.pull()
                    while view is not try_view:
                        view = self.view_stack.pull()
                    self.exception_stack.push(e)
                else:
                    raise e
                
        


from .precedence import PrecedenceGraph
from .core_types import Stack, StretchTerminate, Promise, CommandInterface, Function, Class
from garnish import garnish
import os, importlib
from .builtin_extensions import get_builtin

precedence = PrecedenceGraph()
operators = {}

commands = CommandInterface()

def add_group(group, low=None, high=None):
    precedence.group(group, low, high)
    
@garnish
def add_operation(operate, group:str, symbol:str):
    precedence.add_members(group, symbol)
    operators[symbol] = operate
    return operate

@garnish
def add_command(behavior, *switches):
    commands[switches] = behavior
    return behavior


# OPERATORS
add_group("addition", "comparison")
add_group("multiplication", "addition")
add_group("exponentiation", "multiplication")
add_group("comparison", "equivalence")
add_group("equivalence")

@add_operation.use("addition", "+")
def __add__(l, r):
    return l + r
@add_operation.use("addition", "-")
def __sub__(l, r):
    return l - r
@add_operation.use("multiplication", "*")
def __mul__(l, r):
    return l * r
@add_operation.use("multiplication", "/")
def __div__(l, r):
    return l / r

@add_operation.use("equivalence", "==")
def __eq__(l, r):
    return l == r
@add_operation.use("equivalence", "!=")
def __ne__(l, r):
    return l != r
@add_operation.use("comparison", ">")
def __gt__(l, r):
    return l > r
@add_operation.use("comparison", "<")
def __lt__(l, r):
    return l < r
@add_operation.use("comparison", ">=")
def __ge__(l, r):
    return (l > r) or (l == r)
@add_operation.use("comparison", "<=")
def __le__(l, r):
    return (l < r) or (l == r)


# PRINT

@add_command.use("print", "all")
def _print_all(interpreter, view, tokens:Stack):
    print(*tokens)
    tokens.clear()

@add_command.use("print")
def _print(interpreter, view, tokens:Stack):
    print(tokens.pull())
    
# MOVEMENT

@add_command.use("goto")
def goto(interpreter, view, tokens):
    new_line = tokens.pull_if(int)
    if new_line is not None:
        new_line = new_line
    else:
        new_line = 0
    prev_line = view.current_line
    tokens.push(prev_line)
    view.current_line = new_line
    
@add_command.use("Line")
def current_line(interpreter, view, tokens):
    line = view.current_line
    tokens.push(line)

@add_command.use("process")
def process(interpreter, view, tokens:Stack):
    from .parser import parser
    code = tokens.pull_only(str)
    tokens.push(parser.parse(code))


# RUNNING BLOCKS
@add_command.use("run") # initialize and enter a block
def run_block(interpreter, view, tokens:Stack):
    from .lang import Block
    scope = tokens.pull_only(Block)
    interpreter.process_scope(scope)
    interpreter.init_scope(scope)
    tokens.push(Promise(scope))
@add_command.use("run", "init") # initialize a block
def init_block(interpreter, view, tokens:Stack):
    from .lang import Block
    scope = tokens.pull_only(Block)
    interpreter.init_scope(scope)
    tokens.push(Promise(scope))
@add_command.use("run", "enter") # enter a block
def enter_block(interpreter, view, tokens:Stack):
    from .lang import Block
    scope = tokens.pull_only(Block)
    interpreter.process_scope(scope)
    tokens.push(Promise(scope))
@add_command.use("run", "here") # run a block as if it were in the current scope
def run_inline(interpreter, view, tokens:Stack):
    from .lang import Block
    from .scope_types import HereView, InitHereView
    scope = tokens.pull_only(Block)

    enter_view = HereView(view.scope, scope)
    interpreter.view_stack.insert(enter_view)
    init_view = InitHereView(view.scope, scope)
    interpreter.view_stack.insert(init_view)

    tokens.push(Promise(scope))
@add_command.use("run", "init", "here") # initialize a block as if it were in the current scope
def init_inline(interpreter, view, tokens:Stack):
    from .lang import Block
    from .scope_types import InitHereView
    scope = tokens.pull_only(Block)
    veiw = InitHereView(view.scope, scope)
    interpreter.view_stack.insert(veiw)
    tokens.push(Promise(scope))
@add_command.use("run", "enter", "here") # enter a block as if it were in the current scope
def enter_inline(interpreter, view, tokens:Stack):
    from .lang import Block
    from .scope_types import HereView
    scope = tokens.pull_only(Block)
    veiw = HereView(view.scope, scope)
    interpreter.view_stack.insert(veiw)
    tokens.push(Promise(scope))

# FUNCTIONS
@add_command.use("def")
def define(interpreter, view, tokens):
    from .lang import Pointer, Block
    params = tokens.pull_if(Pointer) or ()
    if isinstance(params, Pointer):
        params = params.reference
    if not isinstance(params, tuple):
        params = (params,)
    block = tokens.pull_only(Block)
    tokens.push(Function(block, params))
    
@add_command.use("def", "method")
def define(interpreter, view, tokens):
    from .lang import Pointer, Block
    self = tokens.pull_only(Block)
    params = tokens.pull_if(Pointer) or ()
    if isinstance(params, Pointer):
        params = params.reference
    if not isinstance(params, tuple):
        params = (params,)
    block = tokens.pull_only(Block)
    block.__scope__["self"] = self
    tokens.push(Function(block, params))

@add_command.use("call")
def call(interpreter, view, tokens):
    foo = tokens.pull_only(Function)
    args = []
    for param in foo.params:
        args.append(tokens.pull())
    foo.call(interpreter, args)
        
    tokens.push(Promise(foo))
    
# CLASSES
@add_command.use("class")
def make_class(interpreter, view, tokens):
    from .lang import Block
    
    class_body = tokens.pull_only(Block)
    from .scope_types import View, InitView
    interpreter.view_stack.push(View(class_body))
    interpreter.view_stack.push(InitView(class_body))
    tokens.push(Class(class_body))

@add_command.use("class", "instance")
def new_instance(interpreter, view, tokens):
    cls = tokens.pull_only(Class)
    tokens.push(cls.new(interpreter, tokens))
    
@add_command.use("class", "extend")
def extend_class(interpreter, view, tokens):
    from .lang import Block
    from .scope_types import View, InitView
    cls = tokens.pull_only(Class)
    class_body = tokens.pull_only(Block)
    class_body.__scope__["super"] = cls.block
    interpreter.view_stack.push(View(class_body))
    interpreter.view_stack.push(InitView(class_body))
    tokens.push(Class(class_body))
    

    
@add_command.use("bind")
def bind(interpreter, view, tokens):
    from .lang import Pointer, Block
    block = tokens.pull_only(Block)
    properties = tokens.pull_only(Pointer).reference
    if not isinstance(properties, tuple):
        properties = (properties,)
    values = tokens.pull()
    if not isinstance(values, tuple):
        values = (values,)
    if len(properties) != len(values):
        raise StretchTerminate()
    scope_dict = block.__scope__
    for index in range(len(properties)):
        scope_dict[properties[index].literal] = values[index]
    
    
# TERMINATING BLOCKS
@add_command.use("break")
def terminate_program(interpreter, view, tokens:Stack):
    interpreter.view_stack.pull()
@add_command.use("break", "count")
def remove_scope(interpreter, view, tokens:Stack):
    count = tokens.pull_if(int) or 1
    while count > 0:
        interpreter.view_stack.pull()
        count -= 1
@add_command.use("break", "all")
def remove_scope(interpreter, view, tokens:Stack):
    interpreter.view_stack = Stack()
@add_command.use("end")
def terminate_program(interpreter, view, tokens:Stack):
    interpreter.running = False  
@add_command.use("continue")
def finish_scope(interpreter, view, tokens:Stack):
    view.current_line = view.scope.end    
@add_command.use("return")
def return_val(interpreter, view, tokens:Stack):
    val = tokens.peek()
    if val is not None:
        view.scope.return_stack.push(val)
    interpreter.view_stack.pull()
@add_command.use("return", "all")
def return_val(interpreter, view, tokens:Stack):
    while len(tokens) > 0:
        view.scope.return_stack.push(tokens.pull())
    interpreter.view_stack.pull()

# CONTROL FLOW
@add_command.use("if")
def if_stat(interpreter, view, tokens:Stack):
    from .lang import Block, RawToken
    condition = tokens.pull_only(bool)
    if condition:
        block = tokens.pull_only(Block)
        from .scope_types import HereView, InitHereView
        enter_view = HereView(view.scope, block)
        interpreter.view_stack.insert(enter_view)
        init_view = InitHereView(view.scope, block)
        interpreter.view_stack.insert(init_view)
        tokens.push(Promise(block))
        return
    tokens.push(None)
@add_command.use("unless")
def unless_stat(interpreter, view, tokens:Stack):
    from .lang import Block, RawToken
    condition = tokens.pull_only(bool)
    if not condition:
        block = tokens.pull_only(Block)
        from .scope_types import HereView, InitHereView
        enter_view = HereView(view.scope, block)
        interpreter.view_stack.insert(enter_view)
        init_view = InitHereView(view.scope, block)
        interpreter.view_stack.insert(init_view)
        tokens.push(Promise(block))
        return
    tokens.push(None)

@add_command.use("True")
def return_true(interpreter, view, tokens:Stack):
    tokens.insert(True)
@add_command.use("False")
def return_false(interpreter, view, tokens:Stack):
    tokens.insert(False)
@add_command.use("None")
def return_none(interpreter, view, tokens:Stack):
    tokens.insert(None)

# EXCEPTION HANDLING
@add_command.use("try")
def _try(interpreter, view, tokens:Stack):
    from .lang import Block
    from .scope_types import HereView, InitHereView
    scope = tokens.pull_only(Block)
    scope_view = HereView(view.scope, scope)
    init_scope_view = InitHereView(view.scope, scope)
    interpreter.view_stack.push(scope_view)
    interpreter.view_stack.push(init_scope_view)
    interpreter.try_stack.push(scope_view)
    tokens.push(Promise(scope))
@add_command.use("catch")
def catch(interpreter, view, tokens:Stack):
    from .lang import Block, RawToken
    exc = interpreter.exception_stack.pull_if(Exception)
    if exc is not None:
        scope = tokens.pull_if(Block)
        if scope is not None:
            scope.__scope__["EXCEPTION"] = exc
            interpreter.run_scope(scope)
            tokens.push(Promise(scope))
            return
    tokens.push(None)
@add_command.use("raise")
def make_exception(interpreter, view, tokens:Stack):
    message = tokens.pull_if(str)
    if message is None:
        message = "raised without an exception message!"
    message = f"[Line : {view.current_line}] " + message
    raise StretchTerminate(message)



@add_command.use("import")
def import_module(interpreter, view, tokens:Stack):
    from .lang import Pointer
    pointer = tokens.pull_only(Pointer)
    dotpath = pointer.reference
    dot = dotpath.dot()
    if dot in interpreter.__modules__:
        scope = interpreter.__modules__[dot]
    else:
        filepath = dotpath.file("str")
        if os.path.exists(filepath):
            from .parser import parser
            with open(filepath, 'r') as module_file:
                scope = parser.parse(module_file.read())
            interpreter.__modules__[dot] = scope
        else:
            raise StretchTerminate(f"There is no module {filepath}")
    tokens.insert(scope)
@add_command.use("extend")
def import_extension(interpreter, view, tokens:Stack):
    from .lang import Pointer
    pointer = tokens.pull_only(Pointer)
    path = pointer.reference
    dot = path.dot()
    if pointer.marked:
        if dot in interpreter.__builtin__:
            extender = interpreter.__builtin__[dot]
        else:
            extender = get_builtin(path)
        extender.extend(view.scope)
    else:
        if dot in interpreter.__extensions__:
            extender = interpreter.__extensions__[dot]
        else:
            filepath = path.file("py")
            if os.path.exists(filepath):
                extender = importlib.import_module(dot).__extension__
                interpreter.__extensions__[dot] = extender
            else:
                raise StretchTerminate(f"There is no extension {filepath}")
        extender.extend(view.scope)

# COMMUNICATION
# global stack
@add_command.use("stack", "push")
def push_stack(interpreter, view, tokens:Stack):
    interpreter.stack.push(tokens.pull())
@add_command.use("stack", "pull")
def pull_stack(interpreter, view, tokens:Stack):
    val = interpreter.stack.pull()
    if isinstance(val, Promise):
        val = val.load()
    tokens.push(val)
@add_command.use("stack", "look")
def pull_stack(interpreter, view, tokens:Stack):
    val = interpreter.stack.peek()
    if isinstance(val, Promise):
        val = val.load()
    tokens.push(val)
# global channels
@add_command.use("channel", "emit")
def emit_channel(interpreter, view, tokens:Stack):
    from .lang import Pointer
    ids = tokens.pull_only(Pointer).reference
    match ids:
        case ids if isinstance(ids, tuple):
            vals = tokens.pull()
            if not isinstance(vals, tuple):
                vals = tuple([vals for _ in range(len(ids))])
            for index in range(len(ids)):
                interpreter.channels.emit(ids[index], vals[index])
        case id:
            interpreter.channels.emit(id, tokens.pull())
                
    
@add_command.use("channel", "look")
def recieve_channel(interpreter, view, tokens:Stack):
    from .lang import Pointer
    match tokens.pull_only(Pointer).reference:
        case ids if isinstance(ids, tuple):
            for id in ids:
                tokens.push(interpreter.channels.receive(id))
        case id:
            tokens.push(interpreter.channels.receive(id))

@add_command.use("channel", "take")
def take_channel(interpreter, view, tokens:Stack):
    from .lang import Pointer
    ids = tokens.pull_only(Pointer).reference
    if isinstance(ids, tuple):
        result = []
        for id in ids:
            result.append(interpreter.channels.take(id))
        tokens.push(tuple(result))
        return
    tokens.push(interpreter.channels.take(ids))

@add_command.use("channel", "on")
def on_channel(interpreter, view, tokens:Stack):
    from .lang import RawToken, Block, Pointer
    ids = tokens.pull_only(Pointer).reference
    block = tokens.pull_only(Block)
    if not isinstance(ids, tuple):
        ids = (ids,)
    for id in ids:
        if interpreter.channels.receive(id.literal) is None:
            return
    from .scope_types import HereView, InitHereView
    enter_view = HereView(view.scope, block)
    interpreter.view_stack.insert(enter_view)
    init_view = InitHereView(view.scope, block)
    interpreter.view_stack.insert(init_view)
    tokens.push(block)

@add_command.use("open")
def open_pointer(interpreter, view, tokens:Stack):
    from .lang import Pointer, Expression
    point = tokens.pull_only(Pointer)
    val = point.reference
    tokens.insert(view.scope.process_expr(interpreter, view, val))

@add_command.use("Scope")
def get_scope(interpreter, view, tokens:Stack):
    tokens.insert(view.scope)

        


class Stack(list):
    __slots__ = ("single_layer")
    class StackError(Exception): pass
    def __init__(self, *iterable, single_layer=True):
        if single_layer:
            if any([isinstance(obj, Stack) and obj is not self for obj in iterable]):
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        self.single_layer = single_layer
        super().__init__(iterable)
        
    def __repr__(self):
        return f"Stack -> {list.__repr__(self)}"    
    
    def level(self, index = 1):
        if self.single_layer:
            if index != 1:
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
            return self
        current_stack = self
        stacks = [current_stack]
        while len(current_stack) > 0:
            if isinstance(list.__getitem__(current_stack, -1), Stack):
                current_stack = current_stack[-1]
                stacks.append(current_stack)
            else:
                break
        return stacks[-index]
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return list.__getitem__(self.level(), key)
        elif isinstance(key, slice):
            return Stack(*list.__getitem__(self.level(), key), single_layer=self.single_layer)
        
    def up(self, levels:int=1):
        if self.single_layer:
            raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        for _ in range(levels):
            self.level().push(Stack())
    def down(self, levels:int = 1):
        if self.single_layer:
            raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        return self.level(1 + levels).pop()
    
    def push(self, value):
        if self.single_layer:
            if isinstance(value, Stack) and value is not self:
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        self.level().insert(value, 0)
    def pull(self, index=0):
        return self.level().pop(index)

    def pull_if(self, type, index=0):
        val = self.peek(index)
        if isinstance(val, type):
            self.pull(index)
            return val
        else:
            return None
    def pull_only(self, type, index=0):
        val = self.pull_if(type, index)
        if val is None:
            raise Stack.StackError(f"expected {type} at {index} on stack recieved {self.peek(index)}")
        return val
    
    def peek(self, index=0):
        if index > len(self.level()) - 1:
            return None
        return self.level()[index]

    def peek_if(self, type:type, index=0):
        val = self.peek(index)
        if not isinstance(val, type):
            return None
        return val
    def empty(self, level:int=1):
        self.level(level).clear()
    def copy(self, index = 0):
        self.push(self.peek(index))
    def surface(self, index=-1):
        self.push(self.pull(index))

    def insert(self, value, index=0):
        list.insert(self.level(), index, value)
class Channel:
    def __init__(self):
        self.channels = {}
        
    def emit(self, id, value):
        self.channels[id] = value
        
    def clear(self):
        self.channels.clear()
        
    def receive(self, id, default=None):
        value = default
        if id in self.channels:
            value = self.channels[id]
        return value
    
    def take(self, id, default=None):
        value = default
        if id in self.channels:
            value = self.channels[id]
            del self.channels[id]
        return value
class StretchTerminate(Exception):
    def __init__(self, *objects):
        message = ""
        for object in objects:
            message += str(object)
        if message == "":
            message = "Exception with no defined message"
        super().__init__(("\x1b[31m" + message + "\x1b[0m").encode().decode())
        
class Promise:
    def __init__(self, block):
        self.block = block
    
    def load(self):
        if len(self.block.return_stack) > 0:
            return self.block.return_stack.pull()
        else:
            return self.block
        
class CommandInterface:
    def __init__(self):
        self.commands = {}
    def __setitem__(self, keys, command):
        index = 0
        current_dict = self.commands
        while index < len(keys):
            key = keys[index]
            if not dict.__contains__(current_dict, key):
                dict.__setitem__(current_dict, key, {'.':None})
            current_dict = dict.__getitem__(current_dict, key)
            index += 1
        dict.__setitem__(current_dict, '.', command)
    
    def __deepcopy__(self, memo):
        return self
    
    def __getitem__(self, command):
        command, switches = command
        if command not in self.commands:
            print(self)
            raise StretchTerminate(f"There is no registered command {command}.")
        current_dict = self.commands[command]
        prev_switch = command
        
        for switch in switches:
            if switch not in current_dict:
                raise StretchTerminate(f"There is no switch '{switch}' registered to \\{command}.")
            current_dict = current_dict[switch]
            prev_switch = switch
        return current_dict['.']
    
    def update(self, other):
        def follow(level, node):
            for key, value in node.items():
                if key == ".":
                    self[level] = value
                else:
                    follow(level + (key,), value)
        follow((), other.commands)
    def copy(self):
        return self
    
class Function:
    def __init__(self, block, params):
        self.block = block
        self.params = params
    
    @property
    def return_stack(self):
        return self.block.return_stack
    
    def call(self, interpreter, args):
        args_length, params_length =len(args), len(self.params)
        if args_length != params_length:
            raise StretchTerminate(f"Function expected {params_length} arguments, recieved {args_length}")
        for index, name in enumerate(self.params):
            self.block.set_var(name, args[index])
        from .scope_types import View, InitView
        interpreter.view_stack.push(View(self.block))
        interpreter.view_stack.push(InitView(self.block))

class Class:
    def __init__(self, block):
        self.block = block
            
    
    def new(self, interpreter, tokens):
        from .lang import Block
        from .scope_types import View, InitView
        from .core import call
        if "__new__" in self.block.__scope__:
            __new__ = self.block.__scope__["__new__"]
            __new__.block.__scope__["cls"] = self.block
            call(interpreter, 0, Stack(__new__, *tokens))
            return Promise(__new__)
        else:
            instance = Block({}, {}, 0)
            instance.__scope__["__class__"] = self.block
            return instance
            
    
    def extend(self, block, params):
        return Class(block, params or self.params, self)
        


from .core_types import Stack
from copy import deepcopy
class View:
    def __init__(self, scope):
        self._scope = scope
        scope.run_counter += 1
        self._here = scope
        self.current_line = 0
        
    @property
    def scope(self):
        return self._scope
    @property
    def here(self):
        return self._here
    
    def step(self, interpreter):
        if self.current_line >= self.scope.end:
            if hasattr(self.scope, '__end__'):
                return self.scope.__end__(interpreter, self) or False
            return False
        
        
        if self.current_line in self.scope.stat:
            statement = self.scope.stat[self.current_line]
            self.here.process_section(interpreter, self, Stack(*statement))
        if self.current_line == 0:
            if hasattr(self.scope, '__start__'):
                self.current_line += 1
                return self.scope.__start__(interpreter, self) or True
        self.current_line += 1
        return True
    def get_var(self, key):
        return self.here.get_var(key)
    def set_var(self, key, value):
        return self.here.set_var(key, value)
class InitView(View):
    def step(self, interpreter):
        if self.current_line >= self.scope.end:
            return False
        
        if self.current_line in self.scope.init:
            statement = self.scope.init[self.current_line]
            self.here.process_section(interpreter, self, Stack(*statement))
        self.current_line += 1
        return True

class InitHereView(InitView):
    def __init__(self, here, scope):
        super().__init__(scope)
        self._here = here
        
    def get_var(self, key):
        return self.here.get_var(key)
    def set_var(self, key, value):
        self.here.set_var(key, value)
class HereView(View):
    def __init__(self, here, scope):
        super().__init__(scope)
        self._here = here
    def get_var(self, key):
        return self.here.get_var(key)
    def set_var(self, key, value):
        self.here.set_var(key, value)




from .precedence import PrecedenceGraph
from .core_types import CommandInterface
from garnish import garnish

class Extender:
    
    def __init__(self):
        self.operators = {}
        self.commands = CommandInterface()
        self.precedence = PrecedenceGraph()
        
        self.direct = []
    
    def extend(self, scope):
        scope.operators.update(self.operators)
        scope.commands.update(self.commands)
        scope.precedence.update(self.precedence)
        
        for func in self.direct:
            func(scope)
    def __enter__(self):
        return self
    def __exit__(*args):
        pass
    
    
    def __call__(self, func):
        self.direct.append(func)
    
    def add_group(self, name:str, low:str=None, high:str=None):
        self.precedence.group(name, low, high)
    
    @garnish
    def operator(self, func, group:str, symbol:str):
        self.precedence.add_members(group, symbol)
        self.operators[symbol] = func
    
    @garnish
    def command(self, func, *switches):
        self.commands[*switches] = func
    
    
        


from stretch.extension import Extender

__extension__ = Extender()
with __extension__ as e:
    @e.command.use("in")
    def get_input(interpreter, view, tokens):
        message = tokens.pull_if(str) or ""
        tokens.insert(input(message))
    
    @e.command.use("ins")
    def get_input(interpreter, view, tokens):
        sentinal = tokens.pull_only(str)
        message = tokens.pull_if(str) or ""
        result = ""
        inp = input(message)
        while inp != sentinal:
            result += inp.encode().decode()
            inp = input(message)
        tokens.insert(result)


from .core import precedence, operators, commands
from .core_types import StretchTerminate, Stack, Promise

class SetVar:
    __slots__ = ("token",)
    def __init__(self, token:'RawToken'):
        if isinstance(token, str):
            token = RawToken(token)
        self.token = token

class RawToken:
    __slots__ = ("literal",)
    __raw__ = {}
    def __init_subclass__(cls):
        cls.__raw__ = dict()
    
    def __deepcopy__(self, memo):
        return self
    
    def __new__(cls, literal:str):
        if issubclass(type(literal), RawToken):
            literal = literal.literal

        if literal in cls.__raw__:
            instance = cls.__raw__[literal]
        else:
            instance = super().__new__(cls)
            instance.literal = literal
            cls.__raw__[literal] = instance
        return instance
    
    def __hash__(self):
        return hash((self.__class__, self.literal))
    
    def __eq__(self, other):
        if other.__class__ is self.__class__:
            return self.literal == other.literal
        return False
    def __repr__(self):
        return f"({type(self).__name__}:'{self.literal}')"
    
    def dot(self):
        return self.literal
    def file(self, ext=None):
        if ext is not None: ext = "." + ext
        return self.literal + ext or ""
    

class Op(RawToken): pass

class Expression:
    def __init__(self, expr):
        self.expr = expr
    
    def load_vars(self, view):
        index = 0
        copy = self.expr[:]
        index = 0
        while index < len(copy):
            val = copy[index]
            if not isinstance(val, Op):
                if isinstance(val, RawToken):
                    copy[index] = view.get_var(val)
                if isinstance(val, Dotpath):
                    copy[index] = view.get_var(val)
                elif isinstance(val, Expression):
                    copy[index] = val.load_vars(view)
            
            index += 1
        return Expression(copy)
    def resolve(self, precedence, operations, solve):
        solution = solve.expr[:]
        for group in precedence:
            index = 0
            while index < len(solution) - 1:
                
                term = solution[index]
                if isinstance(term, Op):
                    op = term.literal
                    if op in group:
                        if op in operators:
                            index -= 1
                            l = solution.pop(index)
                            _ = solution.pop(index)
                            r = solution.pop(index)
                            if isinstance(l, Expression):
                                l = l.resolve(precedence, operations, l)
                            if isinstance(r, Expression):
                                r = r.resolve(precedence, operations, r)
                            value = operators[op](l, r)
                            solution.insert(index, value)
                        else:
                            raise StretchTerminate(f"Undefined Operator {op}")
                index += 1
        if len(solution) > 1:
            undefined_ops = []
            for term in solution:
                if isinstance(term, Op):
                    undefined_ops.append(term)
            raise StretchTerminate(f"Unresolved expression, undefined operators: {undefined_ops} in {solution}")
        return solution[0]
class Dummy:
    def __init__(self, expr):
        self.value = expr
    def load(self):
        from copy import deepcopy
        return deepcopy(self.value)
        

        
class Switch(RawToken): pass
class Command:
    __slots__ = ("command", "switches")
    def __init__(self, command:str, *switches:str):
        self.command = command
        self.switches = switches
    
    def __iter__(self):
        yield self.command
        yield self.switches
    
    def __repr__(self):
        return f"{self.command}: @{' @'.join(self.switches)}>"
                              
class Dotpath:
    __slots__ = ("chain",)
    def __init__(self, *chain):
        self.chain = chain      
    
    def file(self, ext = None):
        if ext is not None: ext = "." + ext
        return "/".join([str(seg.literal) if isinstance(seg, RawToken) else str(seg) for seg in self.chain]) + (ext or "")
    def dot(self):
        return ".".join([str(seg.literal) if isinstance(seg, RawToken) else str(seg) for seg in self.chain])
    
    
    def __repr__(self):
        return f"<{'.'.join([str(seg) for seg in self.chain])}>"

class Pointer:
    __slots__ = ("reference", "marked")
    def __init__(self, obj, mark=False):
        self.marked = mark
        self.reference = obj
    
    def __repr__(self):
        return f"[{self.reference}]"
        

class Statement:
    def __init__(self, sections):
        self.sections = sections
    def __iter__(self):
        for section in self.sections:
            yield section

class Block:
    def __init__(self, stat, init, end):
        self.stat = stat
        self.init = init
        self.end = end
        
        self.run_counter = 0
        
        self.precedence = precedence.copy()
        self.operators = operators.copy()
        self.commands = commands
        
        self.return_stack = Stack()
        self.__scope__ = {}
    
    @property
    def scope(self):
        return self
    
    def set_var(self, keys, values):
        if not isinstance(keys, tuple):
            keys = (keys,)
        if not isinstance(values, tuple):
            values = tuple([values for _ in range(len(keys))])
        for key_index in range(len(keys)):
            set_scope = self.__scope__
            token = keys[key_index]
            if isinstance(token, Dotpath):
                chain = token.chain
                index = 0
                while index < len(chain) - 1:
                    segment = chain[index].literal
                    if segment in set_scope:
                        val = set_scope[segment]
                        if isinstance(val, Block):
                            set_scope = val.__scope__
                            index += 1
                            continue
                    raise StretchTerminate(f"Encountered a gap in {chain} at {segment}")
                else:
                    token = chain[-1]
            name = token.literal
            set_scope[name] = values[key_index]
    def get_var(self, token):        
        get_scope = self.__scope__
        if isinstance(token, Dotpath):
            chain = token.chain
            index = 0
            while index < len(chain) - 1:
                segment = chain[index].literal
                if segment in get_scope:
                    val = get_scope[segment]
                    if isinstance(val, Promise):
                        val = val.load()
                    if isinstance(val, Block):
                        get_scope = val.__scope__
                        index += 1
                        continue
                raise StretchTerminate(f"Encountered a break in {chain} at index {index}: {segment}")
            else:
                token = chain[-1]
                if token.literal not in get_scope:
                    raise StretchTerminate(f"Scope {chain[-2].literal} in chain {chain} does not contain {chain[-1].literal}")
        if token.literal not in get_scope:
            raise StretchTerminate(f"No '{token.literal}' in scope.")
        
        name = token.literal
        value = get_scope[name]
        if isinstance(value, Promise):
            value = value.load()
            get_scope[name] = value
        return value       
    
    def process_section(self, interpreter, view, section):
        index = len(section) - 1
        
        while index >= 0:
            match section[index]:
                case dummy if isinstance(dummy, Dummy):
                    section.pull(index)
                    section.insert(dummy.load(), index)
                case setvar if isinstance(setvar, SetVar):
                    if setvar.token in self.operators:
                        raise StretchTerminate(f"Cannot override term {setvar.token.literal} with a variable.")
                    section.pull(index)
                    view.set_var(setvar.token, section.pull(index))
                case command if isinstance(command, Command):
                    tokens = section[index + 1:]
                    del section[index:]
                    self.commands[command](interpreter, view, tokens)
                    section.extend(tokens)
                case expr if isinstance(expr, (Expression, Dotpath, RawToken, tuple)):
                    section.pull(index)
                    section.insert(self.process_expr(interpreter, view, expr), index)
                
            index -= 1
    def process_expr(self, interpreter, view, expr):
        match expr:
            case expr if isinstance(expr, Expression):
                return expr.resolve(self.precedence.sort_groups(), self.operators, expr.load_vars(view))
            case path if isinstance(path, Dotpath):
                return view.get_var(path)
            case keys if isinstance(keys, tuple):
                result = []
                for key in keys:
                    if key in self.__scope__:
                        result.append(view.get_var(key))
                    else:
                        result.append(key)
                return tuple(result)
            case token if isinstance(token, RawToken):
                if token.literal in self.__scope__:
                    return view.get_var(token)
                else:
                    raise StretchTerminate(f"Unbound token {token}")


    